/**
 * CeboMsr - C++ example - Triggers.
 *
 * This example outlines how two CeboMsr devices can be chained via trigger.
 * The first device is used as master, it is configured to output an alternating trigger,
 * while the second device is configured to respond to this trigger.
 * Both devices sample digital port #0 and single ended input #0.
 * Trigger #0 is used for both master and slave.
 *
 * Requirements: Two CeboLC devices must be connected to the host, GND and TRG must be
 * connected between both devices.
 *
 * This example works with CeboLC only, but can be ported to Cebo-Stick as well.
 *
 * (c) 2013 Cesys GmbH
 * @author Thomas Hoppe
 */

#include <cebomsrpp.h>

#include <stdexcept>
#include <iostream>

using namespace std;
using namespace CeboMsr;

// Cross platform sleep ...
#ifdef	_WIN32
#	include <windows.h>
	inline void cesleep(int ms) { Sleep(ms); }
#else
#	include <unistd.h>
	inline void cesleep(int ms) { usleep(ms * 1000); }
#endif

// Write frames to console.
void dumpFrames(const string &device, const InputFrameVector &frames) {
	for (size_t i = 0; i < frames.size(); ++i) {
		cout << device << ": "
				<< "se#0: " << frames.at(i).getSingleEnded(0) << " V, "
				<< "dp#0: " << frames.at(i).getDigitalPort(0) << endl;
	}
}

// Start DAQ, read frames and output the results.
void runDataAcquisition(Device &master, Device &slave) {
	// The slave must be started first, as it reacts on the master's trigger.
	// Continuous DAQ is used. Timed using an external trigger.
	slave.startContinuousExternalTimedDataAcquisition();

	// The master uses hardware timed DAQ, continuous at 50 Hz.
	// The 'false' here signals that the process should start immediately.
	master.startContinuousDataAcquisition(50, false);

	// The example reads at least 10 frames from 
	// both devices and subsequently output the samples.
	int masterFrames = 0, slaveFrames = 0;
	while (masterFrames < 10 || slaveFrames < 10) {
		// Start by reading frames from master,
		// output it and increment counter.
		InputFrameVector frames = master.readNonBlocking();
		dumpFrames("master", frames);
		masterFrames += (int)frames.size();

		// Do the same with the slave.
		frames = slave.readNonBlocking();
		dumpFrames("slave", frames);
		slaveFrames += (int)frames.size();

		// Don't poll to frequent, this would fully utilize one core.
		cesleep(1);
	}

	// Finished, gracefully stop DAQ.
	slave.stopDataAcquisition();
	master.stopDataAcquisition();
}

// Both devices are fully configured here.
void configure(Device &master, Device &slave) {
	// The trigger for the master must be set to alternating output.
	master.getTriggers().at(0).setConfig(Trigger::OutputAlternating);

	// The slave's trigger must be set to alternating as well.
	slave.getTriggers().at(0).setConfig(Trigger::InputAlternating);

	// Both devices now gets configured to the same input frame layout.
	InputVector inputs;
	inputs
		<< master.getSingleEndedInputs().at(0)
		<< master.getDigitalPorts().at(0);
	master.setupInputFrame(inputs);

	// Use the same list, clear it before use.
	inputs.clear();
	inputs
		<< slave.getSingleEndedInputs().at(0)
		<< slave.getDigitalPorts().at(0);
	slave.setupInputFrame(inputs);
}

// Application entry point.
int main() {
	try {
		// Search for the devices, exactly two are required.
		DeviceVector devices = LibraryInterface::enumerate(DeviceType::CeboLC);
		if (2 != devices.size()) {
			cout << "Exactly two devices are required." << endl;
			return 1;
		}

		// As both devices can act as master, 
		// we simply use the first one for this role.
		Device master = devices[0];
		master.open();
		cout << "Master: " << master.getSerialNumber()
				<< "@" << master.getIdentifier() << endl;

		// ... and the second as slave.
		Device slave = devices[1];
		slave.open();
		cout << "Slave: " << slave.getSerialNumber()
				<< "@" << slave.getIdentifier() << endl;

		// Configure both master and slave ...
		configure(master, slave);

		// ... and do the DAQ.
		runDataAcquisition(master, slave);

		// Close both.
		master.close();
		slave.close();
	} catch (std::exception &ex) {
		cout << "Exception: " << ex.what() << endl;
	}
	return 0;
}

